@@ -685,11 +685,10 @@ gboolean mcap_mdl_abort(struct mcap_mdl *mdl, mcap_mdl_notify_cb abort_cb,
 
 static struct mcap_mcl *find_mcl(GSList *list, const bdaddr_t *addr)
 {
-	GSList *l;
 	struct mcap_mcl *mcl;
 
-	for (l = list; l; l = l->next) {
-		mcl = l->data;
+	for (; list; list = list->next) {
+		mcl = list->data;
 
 		if (!bacmp(&mcl->addr, addr))
 			return mcl;
@@ -792,7 +791,7 @@ static void mcap_cache_mcl(struct mcap_mcl *mcl)
 		last->ctrl &= ~MCAP_CTRL_CACHED;
 		if (last->ctrl & MCAP_CTRL_CONN) {
 			/* We have to release this MCL if */
-			/* connection is not succesful    */
+			/* connection is not successful    */
 			last->ctrl |= MCAP_CTRL_FREE;
 		} else {
 			mcap_mcl_release(last);
@@ -1126,6 +1125,7 @@ static void process_md_abort_mdl_req(struct mcap_mcl *mcl, void *cmd,
 	req = cmd;
 	mdl_id = ntohs(req->mdl);
 	mcl->state = MCL_CONNECTED;
+	abrt = NULL;
 	for (l = mcl->mdls; l; l = l->next) {
 		mdl = l->data;
 		if (mdl_id == mdl->mdlid && mdl->state == MDL_WAITING) {
