@@ -1211,8 +1211,7 @@ int delete_device_service(const bdaddr_t *sba, const bdaddr_t *dba)
 		textfile_del(filename, key);
 	}
 
-	g_slist_foreach(match.keys, (GFunc) g_free, NULL);
-	g_slist_free(match.keys);
+	g_slist_free_full(match.keys, g_free);
 
 	/* Deleting all attributes values of a given address */
 	memset(&match, 0, sizeof(match));
@@ -1228,8 +1227,7 @@ int delete_device_service(const bdaddr_t *sba, const bdaddr_t *dba)
 		textfile_del(filename, key);
 	}
 
-	g_slist_foreach(match.keys, (GFunc) g_free, NULL);
-	g_slist_free(match.keys);
+	g_slist_free_full(match.keys, g_free);
 
 	return 0;
 }
